typedef struct __attribute__((packed)) _tagARGS{
	float canvasrt_top;
    float canvasrt_bottom;
	float canvasrt_left;
	int srcWidth ,srcHeight;
	int dstWidth ,dstHeight;
	int begin_x ,end_x;
	int begin_y ,end_y;
	int dydstx ,dydsty;
	int dxdstx ,dxdsty;
	int int_startdstx ,int_startdsty;
	int linesize;
	int blockOffsetX, blockOffsetY;
	int begin_dstx, end_dstx;
	int begin_dsty ,end_dsty;
	float bgColorRed, bgColorGreen, bgColorBlue;
} ARGS;

typedef struct __attribute__((packed)) _tagCOUNTTABLE{
	ulong pack;
} COUNTTABLE;

#define BLOCK_SIZE 16

__kernel void CRRHFDIBInit8( __global COUNTTABLE * const countTable,
							__global const ARGS * const arg)
{
    int x ,y;
    int id = get_global_id(0) * get_global_size(1) + get_global_id(1);
	int dstWidth = arg[id].dstWidth;

	for( y = arg[id].begin_dsty ; y < arg[id].end_dsty ; y++){
		for( x = arg[id].begin_dstx ; x < arg[id].end_dstx ; x++){
			countTable[ y * dstWidth + x].pack = (ulong)0;
       }
    }
}



__kernel void CRRHFDIBCount8(__global const unsigned char * const canvasPixel,
								__global const COUNTTABLE * const colorTable,
								__global COUNTTABLE * const countTable,
								__global const ARGS * const arg)
{
	float canvasrt_top;
	float canvasrt_left;
    int begin_x;
    int end_x;
	int begin_y;
	int end_y;
	int dydstx, dydsty;
	int int_startdstx, int_startdsty;
	int dstWidth;
    int srcWidth;
    int srcHeight;
    int int_dstx ,int_dsty;
    int dxdstx ,dxdsty;
	int blockOffsetX, blockOffsetY;
    __global const unsigned char *p;

    int x ,y;
    int id;
	
    id = get_global_id(0) * get_global_size(1) + get_global_id(1);
    
	canvasrt_top = arg[id].canvasrt_top;
	canvasrt_left = arg[id].canvasrt_left;
	begin_x = arg[id].begin_x;
    end_x = arg[id].end_x;
	begin_y = arg[id].begin_y;
	end_y = arg[id].end_y;
	dydstx = arg[id].dydstx;
	dydsty = arg[id].dydsty;
	int_startdstx = arg[id].int_startdstx;
	int_startdsty = arg[id].int_startdsty;
	dstWidth = arg[id].dstWidth;
    srcWidth = arg[id].srcWidth;
    srcHeight = arg[id].srcHeight;
    dxdstx = arg[id].dxdstx;
    dxdsty = arg[id].dxdsty;
	blockOffsetX = arg[id].blockOffsetX;
	blockOffsetY = arg[id].blockOffsetY;

	//__private COUNTTABLE write_cache[BLOCK_SIZE * BLOCK_SIZE];

	//Load to write_cache from countTable
	//for (y = 0; y < BLOCK_SIZE; y++){
	//	for (x = 0; x < BLOCK_SIZE; x++){
	//		write_cache[y * BLOCK_SIZE + x] = countTable[(blockOffsetY + y) * dstWidth + (blockOffsetX + x)];
	//	}
	//}

    for( y = begin_y ; y < end_y ; y++){
        int_dstx = int_startdstx + dydstx * (y - canvasrt_top);
		int_dsty = int_startdsty + dydsty * (y - canvasrt_top);
        
        int_dstx += dxdstx * (begin_x - canvasrt_left);
		int_dsty += dxdsty * (begin_x - canvasrt_left);

		p = canvasPixel;
		p += (srcHeight - 1 - y) * ((srcWidth + 3) / 4 * 4) + begin_x;
		#pragma unroll 16
		for(x = begin_x ; x < end_x ; x++){
            int dstx;
            int dsty;
			//__global const COUNTTABLE *color;
			//unsigned char pixel;
			//int2 coord;


            dstx = int_dstx >> 16;
            dsty = int_dsty >> 16;
            


			//color = &colorTable[*p++];
			//countTable[dsty * dstWidth + dstx].pack += color->pack;
			countTable[dsty * dstWidth + dstx].pack += colorTable[*p++].pack;



			//coord.x = x;
			//coord.y = srcHeight-1 -y;
			//pixel = read_imageui(canvasPixel, CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_NONE | CLK_FILTER_NEAREST, coord).x;
			//color = &colorTable[pixel];



			//p++;
			//dstx -= blockOffsetX;
			//dsty -= blockOffsetY;
			//bool cond = 0 <= dstx && dstx < BLOCK_SIZE && 0 <= dsty && dsty < BLOCK_SIZE;
			//if (cond){
			//	color = &colorTable[*p];
			//	write_cache[dsty * BLOCK_SIZE + dstx].pack += color->pack;
			//}




            int_dstx += dxdstx;
            int_dsty += dxdsty;
        }
    }

	//Write to countTable from write_cache
	//for (y = 0; y < BLOCK_SIZE; y++){
	//	for (x = 0; x < BLOCK_SIZE; x++){
	//		countTable[(blockOffsetY + y) * dstWidth + (blockOffsetX + x)] = write_cache[y * BLOCK_SIZE + x];
	//	}
	//}


}

__kernel void CRRHFDIBDiv8( __global COUNTTABLE * const countTable,
							__global unsigned int * const dstPixel,
							__global const ARGS * const arg)
{
    int x ,y;
    int id = get_global_id(0) * get_global_size(1) + get_global_id(1);
	int dstWidth = arg[id].dstWidth;
    int dstHeight = arg[id].dstHeight;
	float bgR, bgG, bgB;

	bgR = arg[id].bgColorRed;
	bgG = arg[id].bgColorGreen;
	bgB = arg[id].bgColorBlue;

	for( y = arg[id].begin_dsty ; y < arg[id].end_dsty ; y++){
		for( x = arg[id].begin_dstx ; x < arg[id].end_dstx ; x++){
			unsigned int pixel;
			unsigned int r, g, b;
			float d;
			float fr, fg, fb;
			ulong pack;

			pack = countTable[(dstHeight - 1 - y) * dstWidth + x].pack;
			r = pack >> 45;
			g = pack >> 26 & (2 << 18) - 1;
			b = pack >> 7 & (2 << 18) - 1;
			d = (pack & 127) * 65536 + 1;
			fr = (r / 4095.0f * 65536.0f + bgR) / d;
			fg = (g / 4095.0f * 65536.0f + bgG) / d;
			fb = (b / 4095.0f * 65536.0f + bgB) / d;

			r = native_powr(fr, 1 / 2.2f) * 255 + .5f;
			g = native_powr(fg, 1 / 2.2f) * 255 + .5f;
			b = native_powr(fb, 1 / 2.2f) * 255 + .5f;

			pixel = ((unsigned int)r << 16) + ((unsigned int)g << 8) + (unsigned int)b;

			dstPixel[y * dstWidth + x] = pixel;

       }
    }
}
